#ifndef __JABYENGINE_CD_INTERNAL_HPP__
#define __JABYENGINE_CD_INTERNAL_HPP__
#include <PSX/System/IOPorts/cd_io.hpp>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            enum struct State {
                Free = 0,
                Done = 0,

                Reading,
                BufferFull,
                Error,
            };

            extern VolatilePOD<CD_IO::Interrupt::Type> last_interrupt;
            extern VolatilePOD<State>                  state;

            struct Sector {
                uint32_t data[512];
            };

            struct FileInfo {
                uint16_t lba;
                uint16_t sectors;

                constexpr FileInfo(uint16_t lba, uint16_t sectors) : lba(lba), sectors(sectors) {
                }
            };

            class SectorBufferAllocator {
            private:
                struct Dummy {
                };

                typedef Sector* (*SimpleAllocator)();
                typedef Sector* (Dummy::*ComplexAllocator)();

            private:
                Dummy* this_ctx = nullptr;
                union {
                    SimpleAllocator allocate;
                    ComplexAllocator this_allocate;
                };

                SectorBufferAllocator(SimpleAllocator allocate) : this_ctx(nullptr), allocate(allocate) {
                }

                SectorBufferAllocator(Dummy* ctx, ComplexAllocator this_allocate) : this_ctx(ctx), this_allocate(this_allocate) {
                }

            public:
                static SectorBufferAllocator create(SimpleAllocator allocate) {
                    return SectorBufferAllocator(allocate);
                }

                template<typename T>
                static SectorBufferAllocator create(T &obj, Sector* (T::*obj_allocate)()) {
                    return SectorBufferAllocator(reinterpret_cast<Dummy*>(&obj), reinterpret_cast<ComplexAllocator>(obj_allocate));
                }
            };

            struct Command {
                static void wait_until(CD_IO::Interrupt::Type irq) {
                    while(last_interrupt.read() != irq);
                }

                template<typename...ARGS>
                static void send(CD_IO::CommandFifo_t& cmd_fifo, CD_IO::ParameterFifo_t& parameter_fifo, CD_IO::Command::Info cmd, ARGS...args) {
                    while(CD_IO::IndexStatus.read().is_bit_set(CD_IO::IndexStatus::IsTransmissionBusy));

                    (parameter_fifo.write(static_cast<uint8_t>(args)), ...);
                    cmd_fifo.write(cmd.id);
                }

                template<typename...ARGS>
                static void send_wait(CD_IO::CommandFifo_t& cmd_fifo, CD_IO::ParameterFifo_t& parameter_fifo, CD_IO::Command::Info cmd, ARGS...args) {
                    send(cmd_fifo, parameter_fifo, cmd, args...);
                    wait_until(cmd.complete_irq);
                }
            };

            State start_reading(FileInfo file_info, const SectorBufferAllocator& buffer_allocator);
        }
    }
}
#endif //!__JABYENGINE_CD_INTERNAL_HPP__